"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class UmapEntry(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    UMAPY_FIELD_NUMBER: builtins.int
    UMAPX_FIELD_NUMBER: builtins.int
    umapY: builtins.float
    umapX: builtins.float
    def __init__(
        self,
        *,
        umapY: builtins.float = ...,
        umapX: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['umapX', b'umapX', 'umapY', b'umapY']) -> None: ...

global___UmapEntry = UmapEntry

@typing.final
class SingleMask(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VERTICES_FIELD_NUMBER: builtins.int
    AREA_FIELD_NUMBER: builtins.int
    TOTALCOUNTS_FIELD_NUMBER: builtins.int
    TOTALGENES_FIELD_NUMBER: builtins.int
    CELLID_FIELD_NUMBER: builtins.int
    CLUSTERID_FIELD_NUMBER: builtins.int
    PROTEINVALUES_FIELD_NUMBER: builtins.int
    NONZEROGENEINDICES_FIELD_NUMBER: builtins.int
    NONZEROGENEVALUES_FIELD_NUMBER: builtins.int
    UMAPVALUES_FIELD_NUMBER: builtins.int
    area: builtins.int
    totalCounts: builtins.int
    totalGenes: builtins.int
    cellId: builtins.str
    clusterId: builtins.str
    @property
    def vertices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def proteinValues(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def nonzeroGeneIndices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def nonzeroGeneValues(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def umapValues(self) -> global___UmapEntry: ...
    def __init__(
        self,
        *,
        vertices: collections.abc.Iterable[builtins.int] | None = ...,
        area: builtins.int = ...,
        totalCounts: builtins.int = ...,
        totalGenes: builtins.int = ...,
        cellId: builtins.str = ...,
        clusterId: builtins.str = ...,
        proteinValues: collections.abc.Iterable[builtins.int] | None = ...,
        nonzeroGeneIndices: collections.abc.Iterable[builtins.int] | None = ...,
        nonzeroGeneValues: collections.abc.Iterable[builtins.int] | None = ...,
        umapValues: global___UmapEntry | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal['umapValues', b'umapValues']) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            'area',
            b'area',
            'cellId',
            b'cellId',
            'clusterId',
            b'clusterId',
            'nonzeroGeneIndices',
            b'nonzeroGeneIndices',
            'nonzeroGeneValues',
            b'nonzeroGeneValues',
            'proteinValues',
            b'proteinValues',
            'totalCounts',
            b'totalCounts',
            'totalGenes',
            b'totalGenes',
            'umapValues',
            b'umapValues',
            'vertices',
            b'vertices',
        ],
    ) -> None: ...

global___SingleMask = SingleMask

@typing.final
class ColormapEntry(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CLUSTERID_FIELD_NUMBER: builtins.int
    COLOR_FIELD_NUMBER: builtins.int
    clusterId: builtins.str
    @property
    def color(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        clusterId: builtins.str = ...,
        color: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['clusterId', b'clusterId', 'color', b'color']) -> None: ...

global___ColormapEntry = ColormapEntry

@typing.final
class Metadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROTEINNAMES_FIELD_NUMBER: builtins.int
    GENENAMES_FIELD_NUMBER: builtins.int
    @property
    def proteinNames(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def geneNames(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        proteinNames: collections.abc.Iterable[builtins.str] | None = ...,
        geneNames: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal['geneNames', b'geneNames', 'proteinNames', b'proteinNames']
    ) -> None: ...

global___Metadata = Metadata

@typing.final
class CellMasks(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CELLMASKS_FIELD_NUMBER: builtins.int
    COLORMAP_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    NUMBEROFCELLS_FIELD_NUMBER: builtins.int
    numberOfCells: builtins.int
    @property
    def cellMasks(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SingleMask]: ...
    @property
    def colormap(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ColormapEntry]: ...
    @property
    def metadata(self) -> global___Metadata: ...
    def __init__(
        self,
        *,
        cellMasks: collections.abc.Iterable[global___SingleMask] | None = ...,
        colormap: collections.abc.Iterable[global___ColormapEntry] | None = ...,
        metadata: global___Metadata | None = ...,
        numberOfCells: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal['metadata', b'metadata']) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            'cellMasks',
            b'cellMasks',
            'colormap',
            b'colormap',
            'metadata',
            b'metadata',
            'numberOfCells',
            b'numberOfCells',
        ],
    ) -> None: ...

global___CellMasks = CellMasks
