"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class PointData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    POSITION_FIELD_NUMBER: builtins.int
    GENENAME_FIELD_NUMBER: builtins.int
    CELLID_FIELD_NUMBER: builtins.int
    geneName: builtins.str
    cellId: builtins.str
    @property
    def position(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        position: collections.abc.Iterable[builtins.float] | None = ...,
        geneName: builtins.str = ...,
        cellId: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal['cellId', b'cellId', 'geneName', b'geneName', 'position', b'position']
    ) -> None: ...

global___PointData = PointData

@typing.final
class TileData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    POINTSDATA_FIELD_NUMBER: builtins.int
    NUMBEROFPOINTS_FIELD_NUMBER: builtins.int
    numberOfPoints: builtins.int
    @property
    def pointsData(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___PointData]: ...
    def __init__(
        self,
        *,
        pointsData: collections.abc.Iterable[global___PointData] | None = ...,
        numberOfPoints: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal['numberOfPoints', b'numberOfPoints', 'pointsData', b'pointsData']
    ) -> None: ...

global___TileData = TileData

@typing.final
class ColumnData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COLUMNTILES_FIELD_NUMBER: builtins.int
    @property
    def columnTiles(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TileData]: ...
    def __init__(
        self,
        *,
        columnTiles: collections.abc.Iterable[global___TileData] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['columnTiles', b'columnTiles']) -> None: ...

global___ColumnData = ColumnData

@typing.final
class LevelData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEVELCOLUMNS_FIELD_NUMBER: builtins.int
    @property
    def levelColumns(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ColumnData]: ...
    def __init__(
        self,
        *,
        levelColumns: collections.abc.Iterable[global___ColumnData] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['levelColumns', b'levelColumns']) -> None: ...

global___LevelData = LevelData

@typing.final
class Transcripts(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEVEL_FIELD_NUMBER: builtins.int
    @property
    def level(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___LevelData]: ...
    def __init__(
        self,
        *,
        level: collections.abc.Iterable[global___LevelData] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['level', b'level']) -> None: ...

global___Transcripts = Transcripts
